import logging
from typing import Optional

from .http1 import TransportHTTP
from .http2 import TransportHTTP2
from .headless import TransportHeadless
from ..fingerprints import get_profile

logger = logging.getLogger("nebulascrape.transport.manager")

TRANSPORT_TIERS = ["http1", "http2", "headless"]


class TransportManager:
    def __init__(self, profile_name: str = "chrome_windows", mode: str = "auto"):
        self.profile_name = profile_name
        self.mode = mode
        self._profile_data = get_profile(profile_name)
        self._current_tier_index: int = 0
        self._success_counts: dict = {t: 0 for t in TRANSPORT_TIERS}
        self._fail_counts: dict = {t: 0 for t in TRANSPORT_TIERS}
        self._active_transport = None
        self._build_transport()

    def _build_transport(self):
        tier = self._current_tier()
        if tier == "http2":
            self._active_transport = TransportHTTP2(self._profile_data)
        elif tier == "headless":
            self._active_transport = TransportHeadless(self.profile_name)
        else:
            self._active_transport = TransportHTTP(self._profile_data)
        logger.debug(f"[TransportManager] Active transport: {tier}")

    def _current_tier(self) -> str:
        if self.mode != "auto":
            return self.mode
        return TRANSPORT_TIERS[min(self._current_tier_index, len(TRANSPORT_TIERS) - 1)]

    def mount_on(self, session):
        self._active_transport.mount_on(session)

    def escalate(self, session):
        if self._current_tier_index < len(TRANSPORT_TIERS) - 1:
            self._current_tier_index += 1
            self._build_transport()
            self._active_transport.mount_on(session)
            logger.info(f"[TransportManager] Escalated to: {self._current_tier()}")
        else:
            logger.warning("[TransportManager] Already at maximum transport tier (headless)")

    def rebuild(self, session):
        self._active_transport.rebuild(session)

    def record_success(self):
        self._success_counts[self._current_tier()] += 1

    def record_failure(self):
        tier = self._current_tier()
        self._fail_counts[tier] += 1
        total = self._fail_counts[tier]
        if total >= 3:
            logger.info(f"[TransportManager] Transport '{tier}' has {total} failures — auto-escalating")
            return True
        return False

    @property
    def current_name(self) -> str:
        return self._current_tier()



import logging
import random
from collections import OrderedDict
from . import BasePlugin
from ..fingerprints import get_profile

logger = logging.getLogger("nebulascrape.plugins.header_optimizer")


class HeaderOptimizerPlugin(BasePlugin):
    name = "header_optimizer"
    priority = 20

    def __init__(self, profile_name: str = "chrome_windows", rotate_ua: bool = False):
        self._profile_name = profile_name
        self._rotate_ua = rotate_ua

    def on_pre_request(self, scraper, method, url, kwargs):
        current_headers = kwargs.get("headers", {}) or {}
        profile = get_profile(self._profile_name)
        base_headers = OrderedDict(profile["headers"])

        merged = OrderedDict()
        for k, v in base_headers.items():
            if k.lower() not in {h.lower() for h in current_headers}:
                merged[k] = v
        merged.update(current_headers)

        if method.upper() in ("POST", "PUT", "PATCH"):
            merged.setdefault("Content-Type", "application/x-www-form-urlencoded")
            merged.pop("Upgrade-Insecure-Requests", None)
            merged["Sec-Fetch-Mode"] = "cors"
            merged["Sec-Fetch-Site"] = "same-origin"

        kwargs["headers"] = merged
        return kwargs



import ssl
import socket
import logging
from requests.adapters import HTTPAdapter

logger = logging.getLogger("nebulascrape.transport.http1")


class TLSFingerprintAdapter(HTTPAdapter):
    def __init__(self, profile_data: dict, *args, **kwargs):
        self._profile = profile_data
        self._ssl_context = None
        self._source_address = kwargs.pop("source_address", None)
        self._server_hostname = kwargs.pop("server_hostname", None)
        self._build_ssl_context()
        super().__init__(*args, **kwargs)

    def _build_ssl_context(self):
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.minimum_version = self._profile.get("tls_min", ssl.TLSVersion.TLSv1_2)
        ctx.maximum_version = self._profile.get("tls_max", ssl.TLSVersion.TLSv1_3)

        cipher_list = self._profile.get("cipher_suite", [])
        if cipher_list:
            try:
                ctx.set_ciphers(":".join(cipher_list))
            except ssl.SSLError:
                pass

        ecdh = self._profile.get("ecdh_curve", "prime256v1")
        try:
            ctx.set_ecdh_curve(ecdh)
        except Exception:
            pass

        if self._server_hostname:
            ctx.server_hostname = self._server_hostname
            ctx.check_hostname = False

        ctx.orig_wrap_socket = ctx.wrap_socket
        ctx.wrap_socket = self._wrap_socket_override(ctx)
        self._ssl_context = ctx

    def _wrap_socket_override(self, ctx):
        server_hostname = self._server_hostname

        def wrap_socket(*args, **kwargs):
            if server_hostname:
                kwargs["server_hostname"] = server_hostname
                ctx.check_hostname = False
            else:
                ctx.check_hostname = True
            return ctx.orig_wrap_socket(*args, **kwargs)

        return wrap_socket

    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self._ssl_context
        if self._source_address:
            kwargs["source_address"] = self._source_address
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        kwargs["ssl_context"] = self._ssl_context
        if self._source_address:
            kwargs["source_address"] = self._source_address
        return super().proxy_manager_for(*args, **kwargs)


class TransportHTTP:
    name = "http1"

    def __init__(self, profile_data: dict, **kwargs):
        self.profile_data = profile_data
        self.kwargs = kwargs
        self.adapter = TLSFingerprintAdapter(profile_data, **kwargs)

    def mount_on(self, session):
        session.mount("https://", self.adapter)
        session.mount("http://", HTTPAdapter())

    def rebuild(self, session):
        self.adapter = TLSFingerprintAdapter(self.profile_data, **self.kwargs)
        self.mount_on(session)



class NebulaScrapeError(Exception):
    pass

class CloudflareLoopProtection(NebulaScrapeError):
    pass

class CloudflareIUAMError(NebulaScrapeError):
    pass

class CloudflareSolveError(NebulaScrapeError):
    pass

class CloudflareChallengeError(NebulaScrapeError):
    pass

class CloudflareCaptchaError(NebulaScrapeError):
    pass

class CloudflareCaptchaProvider(NebulaScrapeError):
    pass

class CloudflareCode1020(NebulaScrapeError):
    pass

class TransportError(NebulaScrapeError):
    pass

class FingerprintError(NebulaScrapeError):
    pass

class SessionRotationError(NebulaScrapeError):
    pass

class WAFDetectedError(NebulaScrapeError):
    pass

class TurnstileChallengeError(NebulaScrapeError):
    pass

class RetryExhaustedError(NebulaScrapeError):
    pass



import logging
from .async_client import AsyncNebulaScraper, AsyncClient
import requests
import sys
import ssl
import copyreg

from requests.sessions import Session
from requests.adapters import HTTPAdapter

try:
    from requests_toolbelt.utils import dump as _dump
    _HAS_DUMP = True
except ImportError:
    _HAS_DUMP = False

try:
    import brotli
except ImportError:
    pass

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from .exceptions import (
    CloudflareLoopProtection,
    CloudflareIUAMError,
    NebulaScrapeError,
)
from .cloudflare import Cloudflare
from .user_agent import User_Agent
from .session_intel import SessionIntelligence
from .retry_engine import SmartRetryEngine, RetryAction
from .metrics import MetricsCollector
from .fingerprints import get_profile, available_profiles, DEFAULT_PROFILE
from .transports import TransportManager
from .plugins import PluginRegistry

__version__ = '0.0.1'
__author__ = 'MERO'
__contact__ = 'TG@QP4M'


class NebulaScraper(Session):

    def __init__(self, *args, **kwargs):
        self.debug = kwargs.pop('debug', False)

        self.disableCloudflareV1 = kwargs.pop('disableCloudflareV1', False)
        self.delay = kwargs.pop('delay', None)
        self.captcha = kwargs.pop('captcha', {})
        self.doubleDown = kwargs.pop('doubleDown', True)
        self.interpreter = kwargs.pop('interpreter', 'native')

        self.requestPreHook = kwargs.pop('requestPreHook', None)
        self.requestPostHook = kwargs.pop('requestPostHook', None)

        self.cipherSuite = kwargs.pop('cipherSuite', None)
        self.ecdhCurve = kwargs.pop('ecdhCurve', 'prime256v1')
        self.source_address = kwargs.pop('source_address', None)
        self.server_hostname = kwargs.pop('server_hostname', None)
        self.ssl_context = kwargs.pop('ssl_context', None)

        self.auto_retry = kwargs.pop('auto_retry', False)
        self.max_retries = kwargs.pop('max_retries', 5)

        _profile_name = kwargs.pop('profile', DEFAULT_PROFILE)
        _mode = kwargs.pop('mode', 'auto')

        self.allow_brotli = kwargs.pop(
            'allow_brotli',
            True if 'brotli' in sys.modules.keys() else False
        )

        self.user_agent = User_Agent(
            allow_brotli=self.allow_brotli,
            browser=kwargs.pop('browser', None)
        )

        self._solveDepthCnt = 0
        self.solveDepth = kwargs.pop('solveDepth', 3)

        self._intel = SessionIntelligence()
        self._retry_engine = SmartRetryEngine(max_retries=self.max_retries)
        self._metrics = MetricsCollector()
        self._plugins = PluginRegistry()

        super(NebulaScraper, self).__init__(*args, **kwargs)

        if 'requests' in self.headers.get('User-Agent', ''):
            self.headers = self.user_agent.headers
            if not self.cipherSuite:
                self.cipherSuite = self.user_agent.cipherSuite

        if isinstance(self.cipherSuite, list):
            self.cipherSuite = ':'.join(self.cipherSuite)

        try:
            profile_data = get_profile(_profile_name)
        except ValueError:
            profile_data = get_profile(DEFAULT_PROFILE)

        if _profile_name in available_profiles():
            self._profile_name = _profile_name
        else:
            self._profile_name = DEFAULT_PROFILE

        self._transport_manager = TransportManager(profile_name=self._profile_name, mode=_mode)
        self._transport_manager.mount_on(self)

        copyreg.pickle(ssl.SSLContext, lambda obj: (obj.__class__, (obj.protocol,)))

    def __getstate__(self):
        return self.__dict__

    def perform_request(self, method, url, *args, **kwargs):
        return super(NebulaScraper, self).request(method, url, *args, **kwargs)

    def simpleException(self, exception, msg):
        self._solveDepthCnt = 0
        sys.tracebacklimit = 0
        raise exception(msg)

    @staticmethod
    def debugRequest(req):
        if _HAS_DUMP:
            try:
                print(_dump.dump_all(req).decode('utf-8', errors='backslashreplace'))
            except ValueError as e:
                print(f"Debug Error: {getattr(e, 'message', e)}")

    def decodeBrotli(self, resp):
        if requests.packages.urllib3.__version__ < '1.25.1' and resp.headers.get('Content-Encoding') == 'br':
            if self.allow_brotli and resp._content:
                resp._content = brotli.decompress(resp.content)
            else:
                logging.warning(
                    f'urllib3 {requests.packages.urllib3.__version__} — Brotli content detected '
                    'but allow_brotli=False — skipping decompress.'
                )
        return resp

    def _attach_meta(self, response, intel_result=None, metrics=None):
        if not hasattr(response, 'meta'):
            response.meta = {}
        if intel_result is not None:
            response.meta.update(intel_result.to_meta())
        if metrics is not None:
            response.meta['metrics'] = {
                'latency_ms': metrics.latency_ms,
                'tls_handshake_ms': metrics.tls_handshake_ms,
                'retry_count': metrics.retry_count,
                'redirect_depth': metrics.redirect_depth,
                'transport_used': metrics.transport_used,
            }
        return response

    def register_plugin(self, plugin):
        self._plugins.register(plugin)

    def request(self, method, url, *args, **kwargs):
        if kwargs.get('proxies') and kwargs.get('proxies') != self.proxies:
            self.proxies = kwargs.get('proxies')

        kwargs = self._plugins.run_pre_request(self, method, url, kwargs)

        if self.requestPreHook:
            (method, url, args, kwargs) = self.requestPreHook(self, method, url, *args, **kwargs)

        self._metrics.start(transport=self._transport_manager.current_name)
        self._retry_engine.reset()

        response = None
        last_intel = None

        while True:
            try:
                response = self.decodeBrotli(self.perform_request(method, url, *args, **kwargs))
                self._metrics.mark_connected()
            except Exception as e:
                if self.auto_retry and self._retry_engine.attempt < self.max_retries:
                    self._retry_engine._attempt += 1
                    self._transport_manager.rebuild(self)
                    continue
                raise

            if self.debug:
                self.debugRequest(response)

            last_intel = self._intel.analyze(response)

            if not self.disableCloudflareV1:
                cf_handler = Cloudflare(self)
                if cf_handler.is_Challenge_Request(response):
                    if self._solveDepthCnt >= self.solveDepth:
                        _ = self._solveDepthCnt
                        self.simpleException(
                            CloudflareLoopProtection,
                            f"Loop Protection — solved {_} times in a row."
                        )
                    self._solveDepthCnt += 1
                    self._metrics.set_challenge_solved()
                    response = cf_handler.Challenge_Response(response, **kwargs)
                    last_intel = self._intel.analyze(response)
                else:
                    if not response.is_redirect and response.status_code not in [429, 503]:
                        self._solveDepthCnt = 0

            if self.auto_retry:
                decision = self._retry_engine.analyze_response(response, last_intel)
                self._plugins.run_retry(self, self._retry_engine.attempt, decision)

                if decision.action.value == 'pass':
                    self._transport_manager.record_success()
                    break

                if decision.action.value == 'abort':
                    break

                if decision.rotate_session:
                    try:
                        new_ua = User_Agent(
                            allow_brotli=self.allow_brotli,
                            browser=None
                        )
                        self.headers.update(new_ua.headers)
                    except Exception:
                        pass

                if decision.rebuild_connection:
                    self._transport_manager.rebuild(self)
                elif decision.switch_transport:
                    self._transport_manager.escalate(self)

                self._metrics.increment_retry()
                self._retry_engine.execute_decision(decision)
            else:
                break

        if self.requestPostHook:
            newResponse = self.requestPostHook(self, response)
            if response != newResponse:
                response = newResponse
                if self.debug:
                    print('==== requestPostHook Debug ====')
                    self.debugRequest(response)

        response = self._plugins.run_post_request(self, response, kwargs)

        finished = self._metrics.finish(response)
        self._attach_meta(response, last_intel, finished)

        return response

    @classmethod
    def create_scraper(cls, sess=None, **kwargs):
        scraper = cls(**kwargs)
        if sess:
            for attr in ['auth', 'cert', 'cookies', 'headers', 'hooks', 'params', 'proxies', 'data']:
                val = getattr(sess, attr, None)
                if val is not None:
                    setattr(scraper, attr, val)
        return scraper

    @classmethod
    def get_tokens(cls, url, **kwargs):
        scraper = cls.create_scraper(
            **{
                field: kwargs.pop(field, None) for field in [
                    'allow_brotli', 'browser', 'debug', 'delay',
                    'doubleDown', 'captcha', 'interpreter',
                    'source_address', 'requestPreHook', 'requestPostHook',
                    'profile', 'auto_retry',
                ] if field in kwargs
            }
        )
        try:
            resp = scraper.get(url, **kwargs)
            resp.raise_for_status()
        except Exception:
            logging.error(f'"{url}" returned an error. Could not collect tokens.')
            raise

        domain = urlparse(resp.url).netloc
        cookie_domain = None

        for d in scraper.cookies.list_domains():
            if d.startswith('.') and d in (f'.{domain}'):
                cookie_domain = d
                break
        else:
            cls.simpleException(
                cls,
                CloudflareIUAMError,
                "Unable to find Cloudflare cookies."
            )

        return (
            {'cf_clearance': scraper.cookies.get('cf_clearance', '', domain=cookie_domain)},
            scraper.headers['User-Agent']
        )

    @classmethod
    def get_cookie_string(cls, url, **kwargs):
        tokens, user_agent = cls.get_tokens(url, **kwargs)
        return '; '.join('='.join(pair) for pair in tokens.items()), user_agent


class Client:
    def __init__(
        self,
        profile: str = DEFAULT_PROFILE,
        auto_retry: bool = True,
        max_retries: int = 5,
        mode: str = "auto",
        captcha: dict = None,
        interpreter: str = "native",
        debug: bool = False,
        **kwargs
    ):
        self._scraper = NebulaScraper(
            profile=profile,
            auto_retry=auto_retry,
            max_retries=max_retries,
            mode=mode,
            captcha=captcha or {},
            interpreter=interpreter,
            debug=debug,
            **kwargs
        )

    def get(self, url, **kwargs):
        return self._scraper.get(url, **kwargs)

    def post(self, url, **kwargs):
        return self._scraper.post(url, **kwargs)

    def put(self, url, **kwargs):
        return self._scraper.put(url, **kwargs)

    def delete(self, url, **kwargs):
        return self._scraper.delete(url, **kwargs)

    def request(self, method, url, **kwargs):
        return self._scraper.request(method, url, **kwargs)

    def register_plugin(self, plugin):
        self._scraper.register_plugin(plugin)

    @property
    def metrics(self):
        return self._scraper._metrics.session_metrics()

    @property
    def session(self) -> NebulaScraper:
        return self._scraper


if ssl.OPENSSL_VERSION_INFO < (1, 1, 1):
    print(
        f"WARNING: OpenSSL {ssl.OPENSSL_VERSION} does not support TLS 1.3 — "
        "you may encounter Cloudflare blocks."
    )

create_scraper = NebulaScraper.create_scraper
session = NebulaScraper.create_scraper
get_tokens = NebulaScraper.get_tokens
get_cookie_string = NebulaScraper.get_cookie_string
SmartScraper = Client

__all__ = [
    'NebulaScraper',
    'Client',
    'AsyncNebulaScraper',
    'AsyncClient',
    'SmartScraper',
    'create_scraper',
    'session',
    'get_tokens',
    'get_cookie_string',
    '__version__',
    '__author__',
    '__contact__',
]



import time
from dataclasses import dataclass, field
from typing import Optional


@dataclass
class RequestMetrics:
    latency_ms: float = 0.0
    tls_handshake_ms: float = 0.0
    retry_count: int = 0
    redirect_depth: int = 0
    transport_used: str = "http1"
    status_code: Optional[int] = None
    challenge_solved: bool = False
    waf_vendor: Optional[str] = None


class MetricsCollector:
    def __init__(self):
        self._history: list[RequestMetrics] = []
        self._current: Optional[RequestMetrics] = None
        self._start_time: float = 0.0
        self._connect_time: float = 0.0

    def start(self, transport: str = "http1"):
        self._current = RequestMetrics(transport_used=transport)
        self._start_time = time.monotonic()
        self._connect_time = self._start_time

    def mark_connected(self):
        if self._current is not None:
            elapsed = (time.monotonic() - self._connect_time) * 1000
            self._current.tls_handshake_ms = round(elapsed, 2)

    def increment_retry(self):
        if self._current is not None:
            self._current.retry_count += 1

    def set_redirect_depth(self, depth: int):
        if self._current is not None:
            self._current.redirect_depth = depth

    def set_waf(self, vendor: str):
        if self._current is not None:
            self._current.waf_vendor = vendor

    def set_challenge_solved(self):
        if self._current is not None:
            self._current.challenge_solved = True

    def finish(self, response) -> RequestMetrics:
        if self._current is None:
            return RequestMetrics()
        elapsed = (time.monotonic() - self._start_time) * 1000
        self._current.latency_ms = round(elapsed, 2)
        if response is not None:
            self._current.status_code = response.status_code
            self._current.redirect_depth = len(getattr(response, 'history', []))
        self._history.append(self._current)
        m = self._current
        self._current = None
        return m

    def session_metrics(self) -> dict:
        if not self._history:
            return {}
        latencies = [m.latency_ms for m in self._history]
        return {
            "total_requests": len(self._history),
            "avg_latency_ms": round(sum(latencies) / len(latencies), 2),
            "max_latency_ms": max(latencies),
            "min_latency_ms": min(latencies),
            "total_retries": sum(m.retry_count for m in self._history),
            "challenges_solved": sum(1 for m in self._history if m.challenge_solved),
        }



import logging
import time
import random
from . import BasePlugin

logger = logging.getLogger("nebulascrape.plugins.rate_limit")


class RateLimitPlugin(BasePlugin):
    name = "rate_limit_handler"
    priority = 10

    def __init__(self, min_delay: float = 0.5, max_delay: float = 2.0, burst_threshold: int = 10):
        self._min_delay = min_delay
        self._max_delay = max_delay
        self._burst_threshold = burst_threshold
        self._request_times: list = []

    def _adaptive_delay(self) -> float:
        now = time.monotonic()
        self._request_times = [t for t in self._request_times if now - t < 60]
        count = len(self._request_times)
        if count >= self._burst_threshold:
            return random.uniform(self._max_delay, self._max_delay * 2)
        ratio = count / self._burst_threshold
        return self._min_delay + ratio * (self._max_delay - self._min_delay)

    def on_pre_request(self, scraper, method, url, kwargs):
        delay = self._adaptive_delay()
        if delay > 0.1:
            logger.debug(f"[RateLimitPlugin] Adaptive delay: {delay:.2f}s before {method} {url}")
            time.sleep(delay)
        self._request_times.append(time.monotonic())
        return kwargs

    def on_post_request(self, scraper, response, kwargs):
        if response.status_code == 429:
            ra = response.headers.get("Retry-After")
            if ra:
                try:
                    wait = float(ra) + random.uniform(0.5, 2.0)
                    logger.info(f"[RateLimitPlugin] Retry-After={ra} — sleeping {wait:.1f}s")
                    time.sleep(wait)
                except ValueError:
                    pass
        return response



import logging
from typing import Any, Callable, Dict, List, Optional, Type

logger = logging.getLogger("nebulascrape.plugins")


class BasePlugin:
    name: str = "base"
    priority: int = 50

    def on_pre_request(self, scraper, method: str, url: str, kwargs: dict) -> dict:
        return kwargs

    def on_post_request(self, scraper, response, kwargs: dict):
        return response

    def on_challenge_detected(self, scraper, response, intel_result) -> bool:
        return False

    def on_retry(self, scraper, attempt: int, decision) -> None:
        pass


class PluginRegistry:
    def __init__(self):
        self._plugins: List[BasePlugin] = []

    def register(self, plugin: BasePlugin):
        self._plugins.append(plugin)
        self._plugins.sort(key=lambda p: p.priority)
        logger.debug(f"[PluginRegistry] Registered plugin: {plugin.name}")

    def unregister(self, name: str):
        self._plugins = [p for p in self._plugins if p.name != name]

    def run_pre_request(self, scraper, method, url, kwargs):
        for plugin in self._plugins:
            try:
                kwargs = plugin.on_pre_request(scraper, method, url, kwargs)
            except Exception as e:
                logger.warning(f"[Plugin:{plugin.name}] on_pre_request error: {e}")
        return kwargs

    def run_post_request(self, scraper, response, kwargs):
        for plugin in self._plugins:
            try:
                response = plugin.on_post_request(scraper, response, kwargs)
            except Exception as e:
                logger.warning(f"[Plugin:{plugin.name}] on_post_request error: {e}")
        return response

    def run_challenge(self, scraper, response, intel_result) -> bool:
        handled = False
        for plugin in self._plugins:
            try:
                if plugin.on_challenge_detected(scraper, response, intel_result):
                    handled = True
            except Exception as e:
                logger.warning(f"[Plugin:{plugin.name}] on_challenge_detected error: {e}")
        return handled

    def run_retry(self, scraper, attempt, decision):
        for plugin in self._plugins:
            try:
                plugin.on_retry(scraper, attempt, decision)
            except Exception as e:
                logger.warning(f"[Plugin:{plugin.name}] on_retry error: {e}")



import logging

logger = logging.getLogger("nebulascrape.transport.http2")


class TransportHTTP2:
    name = "http2"

    def __init__(self, profile_data: dict, **kwargs):
        self.profile_data = profile_data
        self.kwargs = kwargs
        self._client = None

    def _build_client(self):
        try:
            import httpx
            headers = dict(self.profile_data.get("headers", {}))
            self._client = httpx.Client(
                http2=True,
                headers=headers,
                verify=True,
                follow_redirects=True,
            )
            return True
        except ImportError:
            logger.warning("[TransportHTTP2] httpx not installed — HTTP/2 unavailable")
            return False

    def mount_on(self, session):
        if self._build_client():
            session._http2_client = self._client
            logger.debug("[TransportHTTP2] HTTP/2 client mounted")
        else:
            from .http1 import TransportHTTP
            transport = TransportHTTP(self.profile_data)
            transport.mount_on(session)

    def request(self, method: str, url: str, **kwargs):
        if self._client is None:
            self._build_client()
        if self._client:
            return self._client.request(method, url, **kwargs)
        raise RuntimeError("HTTP/2 transport not available")

    def rebuild(self, session):
        if self._client:
            try:
                self._client.close()
            except Exception:
                pass
        self._client = None
        self.mount_on(session)



import re
import sys
import time
import random
import hashlib
import logging
from copy import deepcopy
from collections import OrderedDict

try:
    from HTMLParser import HTMLParser
except ImportError:
    if sys.version_info >= (3, 4):
        import html
    else:
        from html.parser import HTMLParser

try:
    from urlparse import urlparse, urljoin
except ImportError:
    from urllib.parse import urlparse, urljoin

from .exceptions import (
    CloudflareCode1020,
    CloudflareIUAMError,
    CloudflareSolveError,
    CloudflareChallengeError,
    CloudflareCaptchaError,
    CloudflareCaptchaProvider,
    TurnstileChallengeError,
    WAFDetectedError,
)
from .captcha import Captcha
from .interpreters import JavaScriptInterpreter

logger = logging.getLogger("nebulascrape.cloudflare")


class Cloudflare:
    def __init__(self, nebulascraper):
        self.nebulascraper = nebulascraper

    @staticmethod
    def unescape(html_text):
        if sys.version_info >= (3, 0):
            if sys.version_info >= (3, 4):
                return html.unescape(html_text)
            return HTMLParser().unescape(html_text)
        return HTMLParser().unescape(html_text)

    @staticmethod
    def is_IUAM_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code in [429, 503]
                and re.search(r'/cdn-cgi/images/trace/jsch/', resp.text, re.M | re.S)
                and re.search(
                    r'<form .*?="challenge-form" action="/\S+__cf_chl_f_tk=',
                    resp.text, re.M | re.S
                )
            )
        except AttributeError:
            pass
        return False

    def is_New_IUAM_Challenge(self, resp):
        try:
            return (
                self.is_IUAM_Challenge(resp)
                and re.search(
                    r'''cpo\.src\s*=\s*['"]/cdn-cgi/challenge-platform/\S+orchestrate/jsch/v1''',
                    resp.text, re.M | re.S
                )
            )
        except AttributeError:
            pass
        return False

    def is_New_Captcha_Challenge(self, resp):
        try:
            return (
                self.is_Captcha_Challenge(resp)
                and re.search(
                    r'''cpo\.src\s*=\s*['"]/cdn-cgi/challenge-platform/\S+orchestrate/(captcha|managed)/v1''',
                    resp.text, re.M | re.S
                )
            )
        except AttributeError:
            pass
        return False

    @staticmethod
    def is_Captcha_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code == 403
                and re.search(r'/cdn-cgi/images/trace/(captcha|managed)/', resp.text, re.M | re.S)
                and re.search(
                    r'<form .*?="challenge-form" action="/\S+__cf_chl_f_tk=',
                    resp.text, re.M | re.S
                )
            )
        except AttributeError:
            pass
        return False

    @staticmethod
    def is_Firewall_Blocked(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code == 403
                and re.search(
                    r'<span class="cf-error-code">1020</span>',
                    resp.text, re.M | re.DOTALL
                )
            )
        except AttributeError:
            pass
        return False

    @staticmethod
    def is_Turnstile_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and re.search(
                    r'cf-turnstile|challenges\.cloudflare\.com/turnstile',
                    resp.text, re.I | re.M
                )
            )
        except AttributeError:
            pass
        return False

    @staticmethod
    def is_Managed_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code in (403, 429, 503)
                and re.search(
                    r'orchestrate/(captcha|managed)/v1',
                    resp.text, re.M | re.S
                )
            )
        except AttributeError:
            pass
        return False

    def is_Challenge_Request(self, resp):
        if self.is_Firewall_Blocked(resp):
            self.nebulascraper.simpleException(
                CloudflareCode1020,
                'Cloudflare has blocked this request (Code 1020 Detected).'
            )

        if self.is_Turnstile_Challenge(resp):
            self.nebulascraper.simpleException(
                TurnstileChallengeError,
                'Cloudflare Turnstile challenge detected — use TransportHeadless with curl_cffi for bypass.'
            )

        if self.is_New_Captcha_Challenge(resp):
            self.nebulascraper.simpleException(
                CloudflareChallengeError,
                'Cloudflare v2 Captcha/Managed challenge detected.'
            )

        if self.is_New_IUAM_Challenge(resp):
            self.nebulascraper.simpleException(
                CloudflareChallengeError,
                'Cloudflare v2 IUAM challenge detected.'
            )

        if self.is_Captcha_Challenge(resp) or self.is_IUAM_Challenge(resp):
            logger.debug('[Cloudflare] Detected v1 challenge.')
            return True

        return False

    def _browser_delay(self, body: str):
        delay = None
        if not self.nebulascraper.delay:
            try:
                delay = float(
                    re.search(r'submit\(\);\r?\n\s*},\s*([0-9]+)', body).group(1)
                ) / 1000.0
            except (AttributeError, ValueError):
                delay = 4.0
            self.nebulascraper.delay = delay

        jitter = random.uniform(-0.2, 0.4)
        sleep_time = max(0, self.nebulascraper.delay + jitter)
        logger.debug(f"[Cloudflare] Browser-realistic delay: {sleep_time:.2f}s")
        time.sleep(sleep_time)

    def IUAM_Challenge_Response(self, body, url, interpreter):
        try:
            formPayload = re.search(
                r'<form (?P<form>.*?="challenge-form" '
                r'action="(?P<challengeUUID>.*?'
                r'__cf_chl_f_tk=\S+)"(.*?)</form>)',
                body, re.M | re.DOTALL
            ).groupdict()

            if not all(key in formPayload for key in ['form', 'challengeUUID']):
                self.nebulascraper.simpleException(
                    CloudflareIUAMError,
                    "Cloudflare IUAM — could not extract form parameters."
                )

            payload = OrderedDict()
            for challengeParam in re.findall(r'^\s*<input\s(.*?)/>', formPayload['form'], re.M | re.S):
                inputPayload = dict(re.findall(r'(\S+)="(\S+)"', challengeParam))
                if inputPayload.get('name') in ['r', 'jschl_vc', 'pass']:
                    payload.update({inputPayload['name']: inputPayload['value']})

        except AttributeError:
            self.nebulascraper.simpleException(
                CloudflareIUAMError,
                "Cloudflare IUAM — could not extract form parameters."
            )

        hostParsed = urlparse(url)
        try:
            payload['jschl_answer'] = JavaScriptInterpreter.dynamicImport(
                interpreter
            ).solveChallenge(body, hostParsed.netloc)
        except Exception as e:
            self.nebulascraper.simpleException(
                CloudflareIUAMError,
                f"Unable to solve Cloudflare challenge: {getattr(e, 'message', e)}"
            )

        return {
            'url': f"{hostParsed.scheme}://{hostParsed.netloc}{self.unescape(formPayload['challengeUUID'])}",
            'data': payload
        }

    def captcha_Challenge_Response(self, provider, provider_params, body, url):
        try:
            formPayload = re.search(
                r'<form (?P<form>.*?="challenge-form" '
                r'action="(?P<challengeUUID>.*?__cf_chl_captcha_tk__=\S+)"(.*?)</form>)',
                body, re.M | re.DOTALL
            ).groupdict()

            if not all(key in formPayload for key in ['form', 'challengeUUID']):
                self.nebulascraper.simpleException(
                    CloudflareCaptchaError,
                    "Cloudflare Captcha — could not extract parameters."
                )

            payload = OrderedDict(
                re.findall(
                    r'(name="r"\svalue|data-ray|data-sitekey|name="cf_captcha_kind"\svalue)="(.*?)"',
                    formPayload['form']
                )
            )
            captchaType = 'reCaptcha' if payload.get('name="cf_captcha_kind" value') == 're' else 'hCaptcha'

        except (AttributeError, KeyError):
            self.nebulascraper.simpleException(
                CloudflareCaptchaError,
                "Cloudflare Captcha — could not extract parameters."
            )

        if self.nebulascraper.proxies and self.nebulascraper.proxies != self.nebulascraper.captcha.get('proxy'):
            self.nebulascraper.captcha['proxy'] = self.nebulascraper.proxies

        self.nebulascraper.captcha['User-Agent'] = self.nebulascraper.headers['User-Agent']

        captchaResponse = Captcha.dynamicImport(provider.lower()).solveCaptcha(
            captchaType, url, payload.get('data-sitekey'), provider_params
        )

        dataPayload = OrderedDict([
            ('r', payload.get('name="r" value', '')),
            ('cf_captcha_kind', payload.get('name="cf_captcha_kind" value', '')),
            ('id', payload.get('data-ray')),
            ('g-recaptcha-response', captchaResponse)
        ])

        if captchaType == 'hCaptcha':
            dataPayload.update({'h-captcha-response': captchaResponse})

        hostParsed = urlparse(url)
        return {
            'url': f"{hostParsed.scheme}://{hostParsed.netloc}{self.unescape(formPayload['challengeUUID'])}",
            'data': dataPayload
        }

    def Challenge_Response(self, resp, **kwargs):
        if self.is_Captcha_Challenge(resp):
            if self.nebulascraper.doubleDown:
                resp = self.nebulascraper.decodeBrotli(
                    self.nebulascraper.perform_request(resp.request.method, resp.url, **kwargs)
                )

            if not self.is_Captcha_Challenge(resp):
                return resp

            if (
                not self.nebulascraper.captcha
                or not isinstance(self.nebulascraper.captcha, dict)
                or not self.nebulascraper.captcha.get('provider')
            ):
                self.nebulascraper.simpleException(
                    CloudflareCaptchaProvider,
                    "Cloudflare Captcha detected — no captcha provider configured."
                )

            if self.nebulascraper.captcha.get('provider') == 'return_response':
                return resp

            submit_url = self.captcha_Challenge_Response(
                self.nebulascraper.captcha.get('provider'),
                self.nebulascraper.captcha,
                resp.text,
                resp.url
            )
        else:
            self._browser_delay(resp.text)

            submit_url = self.IUAM_Challenge_Response(
                resp.text, resp.url, self.nebulascraper.interpreter
            )

        if submit_url:
            def updateAttr(obj, name, newValue):
                try:
                    obj[name].update(newValue)
                    return obj[name]
                except (AttributeError, KeyError):
                    obj[name] = {}
                    obj[name].update(newValue)
                    return obj[name]

            cf_kwargs = deepcopy(kwargs)
            cf_kwargs['allow_redirects'] = False
            cf_kwargs['data'] = updateAttr(cf_kwargs, 'data', submit_url['data'])

            urlParsed = urlparse(resp.url)
            cf_kwargs['headers'] = updateAttr(cf_kwargs, 'headers', {
                'Origin': f'{urlParsed.scheme}://{urlParsed.netloc}',
                'Referer': resp.url
            })

            challengeSubmitResponse = self.nebulascraper.request(
                'POST', submit_url['url'], **cf_kwargs
            )

            if challengeSubmitResponse.status_code == 400:
                self.nebulascraper.simpleException(
                    CloudflareSolveError,
                    'Invalid challenge answer — Cloudflare rejected the response.'
                )

            if not challengeSubmitResponse.is_redirect:
                return challengeSubmitResponse

            cf_kwargs = deepcopy(kwargs)
            cf_kwargs['headers'] = updateAttr(
                cf_kwargs, 'headers', {'Referer': challengeSubmitResponse.url}
            )

            if not urlparse(challengeSubmitResponse.headers['Location']).netloc:
                redirect_location = urljoin(
                    challengeSubmitResponse.url,
                    challengeSubmitResponse.headers['Location']
                )
            else:
                redirect_location = challengeSubmitResponse.headers['Location']

            return self.nebulascraper.request(
                resp.request.method, redirect_location, **cf_kwargs
            )

        return self.nebulascraper.request(resp.request.method, resp.url, **kwargs)



import logging
import random
from typing import List, Optional
from . import BasePlugin

logger = logging.getLogger("nebulascrape.plugins.proxy_manager")


class ProxyManagerPlugin(BasePlugin):
    name = "proxy_manager"
    priority = 30

    def __init__(
        self,
        proxies: List[str] = None,
        rotate_on_fail: bool = True,
        rotate_on_status: List[int] = None,
    ):
        self._proxies = proxies or []
        self._rotate_on_fail = rotate_on_fail
        self._rotate_on_status = rotate_on_status or [403, 429, 503]
        self._current_index = 0
        self._failure_counts: dict = {}

    def _format_proxy(self, proxy_str: str) -> dict:
        return {"http": proxy_str, "https": proxy_str}

    def _current_proxy(self) -> Optional[str]:
        if not self._proxies:
            return None
        return self._proxies[self._current_index % len(self._proxies)]

    def _rotate(self):
        if self._proxies:
            self._current_index = (self._current_index + 1) % len(self._proxies)
            logger.debug(f"[ProxyManager] Rotated to proxy: {self._current_proxy()}")

    def add_proxy(self, proxy: str):
        self._proxies.append(proxy)

    def on_pre_request(self, scraper, method, url, kwargs):
        proxy = self._current_proxy()
        if proxy and "proxies" not in kwargs:
            kwargs["proxies"] = self._format_proxy(proxy)
        return kwargs

    def on_post_request(self, scraper, response, kwargs):
        if response.status_code in self._rotate_on_status and self._rotate_on_fail:
            proxy = self._current_proxy()
            if proxy:
                self._failure_counts[proxy] = self._failure_counts.get(proxy, 0) + 1
                if self._failure_counts[proxy] >= 2:
                    logger.info(f"[ProxyManager] Proxy {proxy} failed twice — rotating")
                    self._rotate()
        elif response.status_code == 200:
            proxy = self._current_proxy()
            if proxy:
                self._failure_counts[proxy] = 0
        return response



import ssl
from collections import OrderedDict
from typing import List


PROFILES = {
    "chrome_windows": {
        "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "headers": OrderedDict([
            ("User-Agent", "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"),
            ("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7"),
            ("Accept-Language", "en-US,en;q=0.9"),
            ("Accept-Encoding", "gzip, deflate, br"),
            ("sec-ch-ua", '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"'),
            ("sec-ch-ua-mobile", "?0"),
            ("sec-ch-ua-platform", '"Windows"'),
            ("Upgrade-Insecure-Requests", "1"),
            ("Sec-Fetch-Dest", "document"),
            ("Sec-Fetch-Mode", "navigate"),
            ("Sec-Fetch-Site", "none"),
            ("Sec-Fetch-User", "?1"),
        ]),
        "cipher_suite": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-CHACHA20-POLY1305",
            "ECDHE-RSA-CHACHA20-POLY1305",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES256-GCM-SHA384",
            "AES128-SHA",
            "AES256-SHA",
        ],
        "ecdh_curve": "prime256v1",
        "tls_min": ssl.TLSVersion.TLSv1_2,
        "tls_max": ssl.TLSVersion.TLSv1_3,
    },
    "chrome_linux": {
        "user_agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "headers": OrderedDict([
            ("User-Agent", "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"),
            ("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7"),
            ("Accept-Language", "en-US,en;q=0.9"),
            ("Accept-Encoding", "gzip, deflate, br"),
            ("sec-ch-ua", '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"'),
            ("sec-ch-ua-mobile", "?0"),
            ("sec-ch-ua-platform", '"Linux"'),
            ("Upgrade-Insecure-Requests", "1"),
            ("Sec-Fetch-Dest", "document"),
            ("Sec-Fetch-Mode", "navigate"),
            ("Sec-Fetch-Site", "none"),
            ("Sec-Fetch-User", "?1"),
        ]),
        "cipher_suite": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-CHACHA20-POLY1305",
            "ECDHE-RSA-CHACHA20-POLY1305",
        ],
        "ecdh_curve": "prime256v1",
        "tls_min": ssl.TLSVersion.TLSv1_2,
        "tls_max": ssl.TLSVersion.TLSv1_3,
    },
    "firefox": {
        "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:121.0) Gecko/20100101 Firefox/121.0",
        "headers": OrderedDict([
            ("User-Agent", "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:121.0) Gecko/20100101 Firefox/121.0"),
            ("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8"),
            ("Accept-Language", "en-US,en;q=0.5"),
            ("Accept-Encoding", "gzip, deflate, br"),
            ("Upgrade-Insecure-Requests", "1"),
            ("Sec-Fetch-Dest", "document"),
            ("Sec-Fetch-Mode", "navigate"),
            ("Sec-Fetch-Site", "none"),
            ("Sec-Fetch-User", "?1"),
            ("TE", "trailers"),
        ]),
        "cipher_suite": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_CHACHA20_POLY1305_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-CHACHA20-POLY1305",
            "ECDHE-RSA-CHACHA20-POLY1305",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES256-GCM-SHA384",
            "AES128-SHA",
            "AES256-SHA",
        ],
        "ecdh_curve": "prime256v1",
        "tls_min": ssl.TLSVersion.TLSv1_2,
        "tls_max": ssl.TLSVersion.TLSv1_3,
    },
    "mobile": {
        "user_agent": "Mozilla/5.0 (Linux; Android 13; Pixel 7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.6099.144 Mobile Safari/537.36",
        "headers": OrderedDict([
            ("User-Agent", "Mozilla/5.0 (Linux; Android 13; Pixel 7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.6099.144 Mobile Safari/537.36"),
            ("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7"),
            ("Accept-Language", "en-US,en;q=0.9"),
            ("Accept-Encoding", "gzip, deflate, br"),
            ("sec-ch-ua", '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"'),
            ("sec-ch-ua-mobile", "?1"),
            ("sec-ch-ua-platform", '"Android"'),
            ("Upgrade-Insecure-Requests", "1"),
            ("Sec-Fetch-Dest", "document"),
            ("Sec-Fetch-Mode", "navigate"),
            ("Sec-Fetch-Site", "none"),
            ("Sec-Fetch-User", "?1"),
        ]),
        "cipher_suite": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-CHACHA20-POLY1305",
            "ECDHE-RSA-CHACHA20-POLY1305",
        ],
        "ecdh_curve": "prime256v1",
        "tls_min": ssl.TLSVersion.TLSv1_2,
        "tls_max": ssl.TLSVersion.TLSv1_3,
    },
}

DEFAULT_PROFILE = "chrome_windows"


def get_profile(name: str) -> dict:
    if name not in PROFILES:
        raise ValueError(f"Unknown fingerprint profile: '{name}'. Available: {list(PROFILES.keys())}")
    return PROFILES[name]


def get_cipher_string(name: str) -> str:
    profile = get_profile(name)
    return ":".join(profile["cipher_suite"])


def available_profiles() -> List[str]:
    return list(PROFILES.keys())



import asyncio
import sys
import logging
from typing import Optional

from .fingerprints import DEFAULT_PROFILE, available_profiles, get_profile
from .session_intel import SessionIntelligence
from .retry_engine import SmartRetryEngine, RetryAction
from .metrics import MetricsCollector
from .plugins import PluginRegistry

logger = logging.getLogger("nebulascrape.async")


class AsyncNebulaScraper:
    def __init__(
        self,
        profile: str = DEFAULT_PROFILE,
        auto_retry: bool = True,
        max_retries: int = 5,
        debug: bool = False,
        **kwargs
    ):
        self._profile_name = profile if profile in available_profiles() else DEFAULT_PROFILE
        self._profile_data = get_profile(self._profile_name)
        self.auto_retry = auto_retry
        self.max_retries = max_retries
        self.debug = debug
        self._extra = kwargs

        self._intel = SessionIntelligence()
        self._retry_engine = SmartRetryEngine(max_retries=max_retries)
        self._metrics = MetricsCollector()
        self._plugins = PluginRegistry()

        self._client = None
        self._closed = False

    def _build_client(self):
        try:
            import httpx
            headers = dict(self._profile_data.get("headers", {}))
            self._client = httpx.AsyncClient(
                http2=True,
                headers=headers,
                verify=True,
                follow_redirects=True,
            )
            return True
        except ImportError:
            logger.warning("[AsyncNebulaScraper] httpx not installed — falling back to aiohttp")
            return False

    def _build_aiohttp_client(self):
        try:
            import aiohttp
            headers = dict(self._profile_data.get("headers", {}))
            connector = aiohttp.TCPConnector(ssl=True, limit=10)
            self._aiohttp_client = aiohttp.ClientSession(
                headers=headers,
                connector=connector,
            )
            return True
        except ImportError:
            logger.warning("[AsyncNebulaScraper] aiohttp not installed either")
            return False

    def _wrap_aiohttp_response(self, resp, body: bytes, url: str):
        class _AiohttpResponseWrapper:
            def __init__(self, r, content, final_url):
                self.status_code = r.status
                self.headers = dict(r.headers)
                self._content = content
                self.url = str(final_url)
                self.history = []
                self.meta = {}
                self.is_redirect = False

            @property
            def text(self):
                return self._content.decode('utf-8', errors='replace')

            @property
            def content(self):
                return self._content

        return _AiohttpResponseWrapper(resp, body, url)

    async def _ensure_client(self):
        if self._client is None:
            if not self._build_client():
                self._build_aiohttp_client()

    async def _do_request(self, method: str, url: str, **kwargs):
        await self._ensure_client()

        if hasattr(self, '_aiohttp_client') and self._client is None:
            async with self._aiohttp_client.request(method, url, **kwargs) as resp:
                body = await resp.read()
                return self._wrap_aiohttp_response(resp, body, resp.url)

        if self._client is not None:
            import httpx
            response = await self._client.request(method, url, **kwargs)
            response.meta = {}
            return response

        raise RuntimeError("No async HTTP backend available — install httpx or aiohttp")

    async def request(self, method: str, url: str, **kwargs) -> object:
        kwargs = self._plugins.run_pre_request(self, method, url, kwargs)
        self._metrics.start(transport="async_http2")
        self._retry_engine.reset()

        response = None
        last_intel = None

        while True:
            try:
                response = await self._do_request(method, url, **kwargs)
                self._metrics.mark_connected()
            except Exception as e:
                if self.auto_retry and self._retry_engine.attempt < self.max_retries:
                    self._retry_engine._attempt += 1
                    await asyncio.sleep(2 ** self._retry_engine.attempt)
                    continue
                raise

            last_intel = self._intel.analyze(response)

            if self.auto_retry:
                decision = self._retry_engine.analyze_response(response, last_intel)
                self._plugins.run_retry(self, self._retry_engine.attempt, decision)

                if decision.action.value in ('pass', 'abort'):
                    break

                self._metrics.increment_retry()
                if decision.backoff_seconds > 0:
                    logger.debug(f"[AsyncNebulaScraper] Retrying in {decision.backoff_seconds:.2f}s — {decision.reason}")
                    await asyncio.sleep(decision.backoff_seconds)
            else:
                break

        response = self._plugins.run_post_request(self, response, kwargs)
        finished = self._metrics.finish(response)

        if not hasattr(response, 'meta'):
            response.meta = {}
        if last_intel is not None:
            response.meta.update(last_intel.to_meta())
        response.meta['metrics'] = {
            'latency_ms': finished.latency_ms,
            'tls_handshake_ms': finished.tls_handshake_ms,
            'retry_count': finished.retry_count,
            'redirect_depth': finished.redirect_depth,
            'transport_used': finished.transport_used,
        }

        return response

    async def get(self, url: str, **kwargs):
        return await self.request("GET", url, **kwargs)

    async def post(self, url: str, **kwargs):
        return await self.request("POST", url, **kwargs)

    async def put(self, url: str, **kwargs):
        return await self.request("PUT", url, **kwargs)

    async def delete(self, url: str, **kwargs):
        return await self.request("DELETE", url, **kwargs)

    def register_plugin(self, plugin):
        self._plugins.register(plugin)

    @property
    def metrics(self):
        return self._metrics.session_metrics()

    async def close(self):
        if self._client:
            await self._client.aclose()
        if hasattr(self, '_aiohttp_client'):
            await self._aiohttp_client.close()
        self._closed = True

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()


AsyncClient = AsyncNebulaScraper



from __future__ import absolute_import

import requests

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from ..exceptions import (
    CaptchaServiceUnavailable,
    CaptchaAPIError,
    CaptchaTimeout,
    CaptchaParameter,
    CaptchaBadJobID
)

try:
    import polling2
except ImportError:
    raise ImportError("Please install the python module 'polling2' via pip")

from . import Captcha


class captchaSolver(Captcha):
    def __init__(self):
        super(captchaSolver, self).__init__('capsolver')
        self.host = 'https://api.capsolver.com'
        self.session = requests.Session()
        self.captchaType = {
            'reCaptcha': 'ReCaptchaV2Task',
            'hCaptcha': 'HCaptchaTask',
            'turnstile': 'AntiCloudflareTask'
        }

    # ------------------------------------------------------------------------------- #

    @staticmethod
    def checkErrorStatus(response, fnct):
        if response.status_code in [500, 502]:
            raise CaptchaServiceUnavailable(f'CapSolver: Server Side Error {response.status_code}')

        try:
            rPayload = response.json()
        except Exception:
            return

        if rPayload.get('errorDescription', False) and 'Current system busy' not in rPayload['errorDescription']:
            raise CaptchaAPIError(
                f"CapSolver -> {fnct} -> {rPayload.get('errorDescription')}"
            )

    # ------------------------------------------------------------------------------- #

    def requestJob(self, jobID):
        if not jobID:
            raise CaptchaBadJobID("CapSolver: Error bad job id to request task result.")

        def _checkRequest(response):
            self.checkErrorStatus(response, 'requestJob')
            try:
                if response.ok and response.json()['status'] == 'ready':
                    return True
            except Exception:
                pass
            return None

        response = polling2.poll(
            lambda: self.session.post(
                f'{self.host}/getTaskResult',
                json={
                    'clientKey': self.api_key,
                    'taskId': jobID
                },
                timeout=30
            ),
            check_success=_checkRequest,
            step=5,
            timeout=180
        )

        if response:
            try:
                rPayload = response.json()['solution']
                if 'token' in rPayload:
                    return rPayload['token']
                else:
                    return rPayload['gRecaptchaResponse']
            except Exception:
                pass

        raise CaptchaTimeout(
            "CapSolver: Error failed to solve Captcha."
        )

    # ------------------------------------------------------------------------------- #

    def requestSolve(self, captchaType, url, siteKey):

        # ------------------------------------------------------------------------------- #

        def _checkRequest(response):
            self.checkErrorStatus(response, 'createTask')
            try:
                rPayload = response.json()
                if response.ok:
                    if rPayload.get("taskId", False):
                        return True
            except Exception:
                pass
            return None

        # ------------------------------------------------------------------------------- #

        payload = {
            'clientKey': self.api_key,
            'appId': '9E717405-8C70-49B3-B277-7C2F2196484B',
            'task': {
                'type': self.captchaType[captchaType],
                'websiteURL': url,
                'websiteKey': siteKey
            }
        }

        if captchaType == 'turnstile':
            payload['task']['metadata'] = {'type': 'turnstile'}

        if self.proxy:
            payload['task']['proxy'] = self.proxy
        else:
            payload['task']['type'] = f"{self.captchaType[captchaType]}Proxyless"

        response = polling2.poll(
            lambda: self.session.post(
                f'{self.host}/createTask',
                json=payload,
                allow_redirects=False,
                timeout=30
            ),
            check_success=_checkRequest,
            step=5,
            timeout=180
        )

        if response:
            rPayload = response.json()
            if rPayload.get('taskId'):
                return rPayload['taskId']

        raise CaptchaBadJobID(
            'CapSolver: Error no job id was returned.'
        )

    # ------------------------------------------------------------------------------- #

    def getCaptchaAnswer(self, captchaType, url, siteKey, captchaParams):
        if not captchaParams.get('api_key'):
            raise CaptchaParameter("CapSolver: Missing api_key parameter.")
        self.api_key = captchaParams.get('api_key')

        if captchaParams.get('proxy') and not captchaParams.get('no_proxy'):
            hostParsed = urlparse(captchaParams.get('proxy', {}).get('https'))

            if not hostParsed.scheme:
                raise CaptchaParameter('Cannot parse proxy correctly, bad scheme')

            if not hostParsed.netloc:
                raise CaptchaParameter('Cannot parse proxy correctly, bad netloc')

            self.proxy = captchaParams['proxy']['https']
        else:
            self.proxy = None

        try:
            jobID = self.requestSolve(captchaType, url, siteKey)
            return self.requestJob(jobID)
        except polling2.TimeoutException:
            raise CaptchaTimeout(
                f"CapSolver: Captcha solve (task ID: {jobID}) took to long."
            )

        raise CaptchaAPIError('CapSolver: Job Failure.')


# ------------------------------------------------------------------------------- #

captchaSolver()



import logging

logger = logging.getLogger("nebulascrape.transport.headless")

_IMPERSONATION_MAP = {
    "chrome_windows": "chrome120",
    "chrome_linux": "chrome120",
    "firefox": "firefox121",
    "mobile": "chrome120",
}


class TransportHeadless:
    name = "headless"

    def __init__(self, profile_name: str = "chrome_windows", **kwargs):
        self.profile_name = profile_name
        self.kwargs = kwargs
        self._session = None

    def _build_session(self):
        try:
            import curl_cffi.requests as curl_req
            impersonate = _IMPERSONATION_MAP.get(self.profile_name, "chrome120")
            self._session = curl_req.Session(impersonate=impersonate)
            logger.debug(f"[TransportHeadless] curl_cffi session built with impersonate={impersonate}")
            return True
        except ImportError:
            logger.warning("[TransportHeadless] curl_cffi not installed — headless transport unavailable")
            return False

    def mount_on(self, session):
        if self._build_session():
            session._headless_client = self._session
            logger.debug("[TransportHeadless] Headless client mounted as fallback")
        else:
            logger.warning("[TransportHeadless] No headless backend available — falling back to HTTP1")
            from .http1 import TransportHTTP
            from ..fingerprints import get_profile
            profile_data = get_profile(self.profile_name)
            t = TransportHTTP(profile_data)
            t.mount_on(session)

    def request(self, method: str, url: str, **kwargs):
        if self._session is None:
            self._build_session()
        if self._session:
            return self._session.request(method, url, **kwargs)
        raise RuntimeError("Headless transport unavailable — install curl_cffi")

    def rebuild(self, session):
        if self._session:
            try:
                self._session.close()
            except Exception:
                pass
        self._session = None
        self.mount_on(session)



import logging
import random
import time
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .exceptions import RetryExhaustedError

logger = logging.getLogger("nebulascrape.retry")


class RetryAction(str, Enum):
    PASS = "pass"
    WAIT_RETRY = "wait_retry"
    ROTATE_SESSION = "rotate_session"
    SWITCH_TRANSPORT = "switch_transport"
    REBUILD_CONNECTION = "rebuild_connection"
    ABORT = "abort"


@dataclass
class RetryDecision:
    action: RetryAction
    backoff_seconds: float = 0.0
    rotate_session: bool = False
    rebuild_connection: bool = False
    switch_transport: bool = False
    reason: str = ""


class SmartRetryEngine:
    def __init__(self, max_retries: int = 5, base_backoff: float = 1.5):
        self.max_retries = max_retries
        self.base_backoff = base_backoff
        self._attempt: int = 0

    def reset(self):
        self._attempt = 0

    @property
    def attempt(self) -> int:
        return self._attempt

    def _jitter(self, value: float, pct: float = 0.25) -> float:
        spread = value * pct
        return value + random.uniform(-spread, spread)

    def _exponential_backoff(self, attempt: int, cap: float = 120.0) -> float:
        raw = self.base_backoff * (2 ** attempt)
        return self._jitter(min(raw, cap))

    def _parse_retry_after(self, resp) -> Optional[float]:
        ra = resp.headers.get("Retry-After")
        if ra is None:
            return None
        try:
            return float(ra)
        except (ValueError, TypeError):
            return None

    def analyze_response(self, resp, intel_result=None) -> RetryDecision:
        if self._attempt >= self.max_retries:
            return RetryDecision(
                action=RetryAction.ABORT,
                reason=f"Max retries ({self.max_retries}) exhausted"
            )

        status = resp.status_code

        if status == 200:
            self._attempt = 0
            return RetryDecision(action=RetryAction.PASS)

        self._attempt += 1

        if status == 429:
            ra = self._parse_retry_after(resp)
            if ra is not None:
                backoff = self._jitter(ra)
            else:
                backoff = self._exponential_backoff(self._attempt, cap=120.0)
            logger.debug(f"[SmartRetry] 429 Rate Limited — waiting {backoff:.1f}s (attempt {self._attempt})")
            return RetryDecision(
                action=RetryAction.WAIT_RETRY,
                backoff_seconds=backoff,
                rotate_session=self._attempt >= 2,
                reason="Rate limited (429)"
            )

        if status == 403:
            backoff = self._jitter(random.uniform(2.0, 8.0))
            logger.debug(f"[SmartRetry] 403 Forbidden — rotating fingerprint+session, waiting {backoff:.1f}s")
            return RetryDecision(
                action=RetryAction.ROTATE_SESSION,
                backoff_seconds=backoff,
                rotate_session=True,
                rebuild_connection=self._attempt >= 3,
                switch_transport=self._attempt >= 4,
                reason="Access denied (403) — rotating fingerprint"
            )

        if status == 503:
            backoff = self._exponential_backoff(self._attempt, cap=60.0)
            logger.debug(f"[SmartRetry] 503 Service Unavailable — rebuilding connection, waiting {backoff:.1f}s")
            return RetryDecision(
                action=RetryAction.REBUILD_CONNECTION,
                backoff_seconds=backoff,
                rebuild_connection=True,
                switch_transport=self._attempt >= 2,
                reason="Service unavailable (503) — rebuilding connection"
            )

        if status in (407, 408, 502, 504, 520, 521, 522, 523, 524, 525, 526):
            backoff = self._exponential_backoff(self._attempt, cap=90.0)
            return RetryDecision(
                action=RetryAction.WAIT_RETRY,
                backoff_seconds=backoff,
                rotate_session=self._attempt >= 2,
                reason=f"Transient error ({status})"
            )

        if intel_result is not None and intel_result.retry_recommended:
            backoff = self._jitter(3.0 + self._attempt * 2)
            return RetryDecision(
                action=RetryAction.ROTATE_SESSION,
                backoff_seconds=backoff,
                rotate_session=intel_result.rotate_session,
                switch_transport=intel_result.switch_transport,
                reason=f"Intel-driven retry: {intel_result.challenge_type}"
            )

        return RetryDecision(action=RetryAction.PASS)

    def execute_decision(self, decision: RetryDecision):
        if decision.action == RetryAction.ABORT:
            raise RetryExhaustedError(decision.reason)

        if decision.backoff_seconds > 0:
            logger.debug(f"[SmartRetry] Sleeping {decision.backoff_seconds:.2f}s — {decision.reason}")
            time.sleep(decision.backoff_seconds)



import os
import re
from setuptools import setup, find_packages
from io import open

with open(os.path.join(os.path.dirname(__file__), 'nebulascrape', '__init__.py')) as fp:
    VERSION = re.match(r".*__version__ = '(.*?)'", fp.read(), re.S).group(1)

with open('README.md', 'r', encoding='utf-8') as fp:
    readme = fp.read()

setup(
    name='nebulascrape',
    author='MERO',
    author_email='mero@ps.com',
    version=VERSION,
    packages=find_packages(exclude=['tests*']),
    description='NebulaScrape — Ultra-powerful HTTP scraping library with smart bypass, async support, and modular transport.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/MERO/nebulascrape',
    keywords=[
        'nebulascrape', 'cloudflare', 'scraping', 'ddos', 'scrape',
        'webscraper', 'anti-bot', 'waf', 'bypass', 'challenge',
        'akamai', 'datadome', 'perimeterx', 'kasada', 'async',
        'fingerprint', 'tls', 'http2'
    ],
    include_package_data=True,
    install_requires=[
        'requests>=2.9.2',
        'requests_toolbelt>=0.9.1',
        'pyparsing>=2.4.7',
        'httpx[http2]>=0.24.0',
        'h2>=4.0.0',
        'aiohttp>=3.8.0',
    ],
    extras_require={
        'headless': ['curl_cffi>=0.5.0'],
        'brotli': ['brotli>=1.0.9'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



import enum
import re
from typing import Optional


class ChallengeType(str, enum.Enum):
    NONE = "none"
    CF_IUAM = "cf_iuam"
    CF_CAPTCHA = "cf_captcha"
    CF_TURNSTILE = "cf_turnstile"
    CF_MANAGED = "cf_managed"
    CF_BLOCK_1020 = "cf_block_1020"
    DATADOME = "datadome"
    PERIMETERX = "perimeterx"
    KASADA = "kasada"
    AKAMAI = "akamai"
    IMPERVA = "imperva"
    SHAPE = "shape"
    JS_REQUIRED = "js_required"
    RATE_LIMITED = "rate_limited"
    REDIRECT_LOOP = "redirect_loop"
    UNKNOWN = "unknown"


class WAFVendor(str, enum.Enum):
    NONE = "none"
    CLOUDFLARE = "cloudflare"
    AKAMAI = "akamai"
    IMPERVA = "imperva"
    DATADOME = "datadome"
    PERIMETERX = "perimeterx"
    KASADA = "kasada"
    SHAPE = "shape"
    UNKNOWN = "unknown"


class IntelResult:
    def __init__(self):
        self.challenge_type: ChallengeType = ChallengeType.NONE
        self.waf_vendor: WAFVendor = WAFVendor.NONE
        self.risk_score: int = 0
        self.retry_recommended: bool = False
        self.rotate_session: bool = False
        self.switch_transport: bool = False
        self.details: dict = {}

    def to_meta(self) -> dict:
        return {
            "challenge_type": self.challenge_type.value,
            "waf_vendor": self.waf_vendor.value,
            "risk_score": self.risk_score,
            "retry_recommended": self.retry_recommended,
            "rotate_session": self.rotate_session,
            "switch_transport": self.switch_transport,
            "details": self.details,
        }


_CF_SERVER_PATTERN = re.compile(r'^cloudflare', re.I)
_CF_IUAM_PATTERN = re.compile(r'/cdn-cgi/images/trace/jsch/', re.M | re.S)
_CF_CHALLENGE_FORM = re.compile(r'<form .*?="challenge-form"', re.M | re.S)
_CF_TURNSTILE = re.compile(r'cf-turnstile|challenges\.cloudflare\.com/turnstile', re.M | re.S | re.I)
_CF_MANAGED = re.compile(r'orchestrate/(captcha|managed)/v1', re.M | re.S)
_CF_JSCH_V1 = re.compile(r'cpo\.src\s*=\s*[\'\"]/cdn-cgi/challenge-platform/\S+orchestrate/jsch/v1', re.M | re.S)
_CF_1020 = re.compile(r'<span class="cf-error-code">1020</span>', re.M | re.DOTALL)
_DATADOME = re.compile(r'datadome\.co|ddg-set-cookie|dd_sitekey', re.I)
_PERIMETERX = re.compile(r'_pxdk|PerimeterX|px\.js|px-captcha|__utmz=px', re.I)
_KASADA = re.compile(r'kasada|kpsdk|k\.js', re.I)
_AKAMAI = re.compile(r'_abck=|ak_bmsc=|akamai-bm-telemetry|sensor_data', re.I)
_IMPERVA = re.compile(r'incap_ses_|visid_incap_|incapsula', re.I)
_SHAPE = re.compile(r'shape\.io|x-shape-|f5\.com/shape', re.I)
_JS_REQUIRED = re.compile(r'<noscript|enable javascript|javascript is required|javascript must be enabled', re.I)
_RETRY_AFTER = re.compile(r'^Retry-After$', re.I)


class SessionIntelligence:
    def __init__(self):
        self._seen_urls: list = []

    def _detect_waf(self, resp) -> WAFVendor:
        server = resp.headers.get('Server', '')
        headers_str = str(resp.headers)
        body = ''
        try:
            body = resp.text[:8192]
        except Exception:
            pass

        if _CF_SERVER_PATTERN.match(server):
            return WAFVendor.CLOUDFLARE
        if _DATADOME.search(headers_str) or _DATADOME.search(body):
            return WAFVendor.DATADOME
        if _PERIMETERX.search(headers_str) or _PERIMETERX.search(body):
            return WAFVendor.PERIMETERX
        if _KASADA.search(headers_str) or _KASADA.search(body):
            return WAFVendor.KASADA
        if _AKAMAI.search(headers_str) or _AKAMAI.search(body):
            return WAFVendor.AKAMAI
        if _IMPERVA.search(headers_str) or _IMPERVA.search(body):
            return WAFVendor.IMPERVA
        if _SHAPE.search(headers_str) or _SHAPE.search(body):
            return WAFVendor.SHAPE
        return WAFVendor.NONE

    def _detect_challenge(self, resp, waf: WAFVendor) -> ChallengeType:
        status = resp.status_code
        body = ''
        try:
            body = resp.text[:16384]
        except Exception:
            pass

        if waf == WAFVendor.CLOUDFLARE:
            if _CF_1020.search(body):
                return ChallengeType.CF_BLOCK_1020
            if _CF_TURNSTILE.search(body):
                return ChallengeType.CF_TURNSTILE
            if _CF_MANAGED.search(body):
                return ChallengeType.CF_MANAGED
            if _CF_JSCH_V1.search(body):
                return ChallengeType.CF_IUAM
            if status in (429, 503) and _CF_IUAM_PATTERN.search(body) and _CF_CHALLENGE_FORM.search(body):
                return ChallengeType.CF_IUAM
            if status == 403 and _CF_CHALLENGE_FORM.search(body):
                return ChallengeType.CF_CAPTCHA

        if waf == WAFVendor.DATADOME and status in (403, 429):
            return ChallengeType.DATADOME
        if waf == WAFVendor.PERIMETERX and status in (403, 429):
            return ChallengeType.PERIMETERX
        if waf == WAFVendor.KASADA and status in (429, 403):
            return ChallengeType.KASADA
        if waf == WAFVendor.AKAMAI and status in (403, 429):
            return ChallengeType.AKAMAI

        if status == 429 or resp.headers.get('Retry-After'):
            return ChallengeType.RATE_LIMITED

        if _JS_REQUIRED.search(body):
            return ChallengeType.JS_REQUIRED

        return ChallengeType.NONE

    def _compute_risk_score(self, status: int, challenge: ChallengeType, waf: WAFVendor) -> int:
        score = 0
        risk_map = {
            ChallengeType.CF_BLOCK_1020: 95,
            ChallengeType.CF_TURNSTILE: 85,
            ChallengeType.CF_MANAGED: 80,
            ChallengeType.CF_CAPTCHA: 75,
            ChallengeType.CF_IUAM: 65,
            ChallengeType.KASADA: 90,
            ChallengeType.PERIMETERX: 80,
            ChallengeType.DATADOME: 75,
            ChallengeType.AKAMAI: 70,
            ChallengeType.IMPERVA: 65,
            ChallengeType.SHAPE: 80,
            ChallengeType.RATE_LIMITED: 40,
            ChallengeType.JS_REQUIRED: 50,
            ChallengeType.NONE: 0,
        }
        score = risk_map.get(challenge, 0)
        if status in (403, 429, 503):
            score = min(score + 10, 100)
        if waf != WAFVendor.NONE and challenge == ChallengeType.NONE:
            score = max(score, 20)
        return score

    def is_redirect_loop(self, resp) -> bool:
        if not hasattr(resp, 'history'):
            return False
        seen = set()
        for r in resp.history:
            if r.url in seen:
                return True
            seen.add(r.url)
        return len(resp.history) > 10

    def analyze(self, resp) -> IntelResult:
        result = IntelResult()

        if self.is_redirect_loop(resp):
            result.challenge_type = ChallengeType.REDIRECT_LOOP
            result.risk_score = 60
            result.retry_recommended = False
            return result

        result.waf_vendor = self._detect_waf(resp)
        result.challenge_type = self._detect_challenge(resp, result.waf_vendor)
        result.risk_score = self._compute_risk_score(resp.status_code, result.challenge_type, result.waf_vendor)

        resolvable = {
            ChallengeType.CF_IUAM, ChallengeType.RATE_LIMITED,
            ChallengeType.DATADOME, ChallengeType.AKAMAI,
        }
        escalate = {
            ChallengeType.CF_TURNSTILE, ChallengeType.CF_MANAGED,
            ChallengeType.CF_CAPTCHA, ChallengeType.KASADA,
            ChallengeType.PERIMETERX, ChallengeType.SHAPE,
        }

        if result.challenge_type in resolvable:
            result.retry_recommended = True
            result.rotate_session = True
        elif result.challenge_type in escalate:
            result.retry_recommended = True
            result.rotate_session = True
            result.switch_transport = True

        result.details["retry_after"] = resp.headers.get("Retry-After")
        result.details["cf_ray"] = resp.headers.get("CF-RAY")
        result.details["status_code"] = resp.status_code

        return result



from .http1 import TransportHTTP
from .http2 import TransportHTTP2
from .headless import TransportHeadless
from .manager import TransportManager

__all__ = [
    "TransportHTTP",
    "TransportHTTP2",
    "TransportHeadless",
    "TransportManager",
]



from .profiles import (
    PROFILES,
    DEFAULT_PROFILE,
    get_profile,
    get_cipher_string,
    available_profiles,
)

__all__ = [
    "PROFILES",
    "DEFAULT_PROFILE",
    "get_profile",
    "get_cipher_string",
    "available_profiles",
]
